import time
from collections import defaultdict
from typing import Any, List, Optional, TypeVar

import sqlalchemy as sa

from bgmi.config import cfg
from bgmi.lib.table import STATUS_UPDATING, Bangumi, Followed, NotFoundError, Session, Subtitle
from bgmi.utils import parse_episode
from bgmi.website.model import Episode, WebsiteBangumi

T = TypeVar("T", bound=dict)


class BaseWebsite:
    parse_episode = staticmethod(parse_episode)

    @staticmethod
    def save_bangumi(data: WebsiteBangumi) -> None:
        """save bangumi to database"""
        with Session.begin() as session:
            subtitle_group = sorted([x.id for x in data.subtitle_group])
            try:
                b = Bangumi.get(Bangumi.id == data.id)

                b.cover = data.cover
                b.update_time = data.update_time
                b.status = STATUS_UPDATING
                b.subtitle_group = subtitle_group

                session.add(b)
            except NotFoundError:
                session.add(
                    Bangumi(
                        name=data.name,
                        id=data.id,
                        update_time=data.update_time,
                        cover=data.cover,
                        status=STATUS_UPDATING,
                        subtitle_group=subtitle_group,
                    )
                )

        for subtitle in data.subtitle_group:
            with Session.begin() as session:
                s = session.scalar(sa.select(Subtitle).where(Subtitle.id == subtitle.id))
                if s:
                    s.name = subtitle.name
                else:
                    session.add(Subtitle(id=subtitle.id, name=subtitle.name))

    def fetch(self, group_by_weekday: bool = True) -> Any:
        bangumi_result = self.fetch_bangumi_calendar()
        if not bangumi_result:
            print("can't fetch anything from website")
            return []

        Bangumi.delete_all()
        for bangumi in bangumi_result:
            self.save_bangumi(bangumi)

        if group_by_weekday:
            result_group_by_weekday = defaultdict(list)
            for bangumi in bangumi_result:
                result_group_by_weekday[bangumi.update_time.lower()].append(bangumi)
            return result_group_by_weekday
        return bangumi_result

    def get_maximum_episode(
        self,
        bangumi: Bangumi,
        ignore_old_row: bool = True,
        max_page: int = cfg.max_path,
    ) -> List[Episode]:
        followed = Followed.get(Followed.bangumi_name == bangumi.name)

        info = self.fetch_single_bangumi(
            bangumi.id,
            subtitle_list=followed.subtitle,
            max_page=max_page,
        )
        if info is not None:
            self.save_bangumi(info)
            data = followed.apply_on_episodes(info.episodes)
        else:
            data = self.fetch_episode_of_bangumi(
                bangumi_id=bangumi.id,
                max_page=max_page,
                subtitle_list=followed.subtitle,
            )
            data = followed.apply_on_episodes(data)

        for episode in data:
            episode.name = bangumi.name

        if ignore_old_row:
            data = [row for row in data if row.time > int(time.time()) - 3600 * 24 * 30 * 3]  # three month

        return data

    def fetch_episode(
        self,
        _id: str,
        name: str = "",
        subtitle_list: Optional[str] = None,
        max_page: int = cfg.max_path,
    ) -> List[Episode]:
        result = []

        max_page = int(max_page)

        if subtitle_list and subtitle_list.split(", "):
            condition = subtitle_list.split(", ")
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, subtitle_list=condition, max_page=max_page)
        else:
            response_data = self.fetch_episode_of_bangumi(bangumi_id=_id, max_page=max_page)

        for info in response_data:
            if "合集" not in info.title:
                info.name = name
                result.append(info)

        return result

    def search_by_keyword(self, keyword: str, count: int) -> List[Episode]:  # pragma: no cover
        """

        :param keyword: search key word
        :param count: how many page to fetch from website
        :return: list of episode search result
        """
        raise NotImplementedError

    def search_by_tag(self, tag: str, subtitle: Optional[str] = None, count: Optional[int] = None) -> List[Episode]:
        """

        :param tag: search tag
        :param subtitle: search subtitle group name
        :param count: how many page to fetch from website
        :return: list of episode search result
        """
        raise NotImplementedError

    def fetch_bangumi_calendar(self) -> List[WebsiteBangumi]:  # pragma: no cover
        """
        return a list of all bangumi and a list of all subtitle group

        list of bangumi dict, update time should be one of
        ['Sun', 'Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Unknown']
        """
        raise NotImplementedError

    def fetch_episode_of_bangumi(
        self, bangumi_id: str, max_page: int, subtitle_list: Optional[List[str]] = None
    ) -> List[Episode]:  # pragma: no cover
        """
        get all episode by bangumi id

        :param bangumi_id: bangumi_id
        :param subtitle_list: list of subtitle group
        :param max_page: how many page to crawl
        :return: list of bangumi
        """
        raise NotImplementedError

    def fetch_single_bangumi(
        self,
        bangumi_id: str,
        subtitle_list: Optional[List[str]] = None,
        max_page: int = cfg.max_path,
    ) -> Optional[WebsiteBangumi]:
        """
        fetch bangumi info when updating, return ``None``
        if website don't have a page contains episodes and info at same time.

        SubClass doesn't have to implement this method.

        :param bangumi_id: bangumi_id, bangumi['keyword']
        :param subtitle_list: list of subtitle group
        :param max_page: how many page to crawl
        """
        return None



import atexit
import pathlib
import pickle

import requests
from requests.adapters import HTTPAdapter, Retry

from bgmi.config import cfg

session = requests.Session()

if cfg.proxy:
    session.proxies = {"http": cfg.proxy, "https": cfg.proxy}

retries = Retry(total=3, backoff_factor=0.1, status_forcelist=[500, 502, 503, 504])
session.mount("https://mikanani.me/", HTTPAdapter(max_retries=retries))

cookies_file = pathlib.Path(cfg.tmp_path).joinpath("mikan_cookies.txt")

if cookies_file.exists():
    data = {}
    dump = cookies_file.read_bytes()
    try:
        data = pickle.loads(dump)
    except pickle.UnpicklingError:
        cookies_file.unlink()
    session.cookies.update(data)


@atexit.register
def save_cookies() -> None:
    if cookies_file.parent.exists() and cookies_file.parent.is_dir():
        cookies_file.write_bytes(pickle.dumps(session.cookies))



import os
import traceback
from typing import List, cast

import stevedore
from stevedore.exception import NoMatches

from bgmi import namespace
from bgmi.config import cfg
from bgmi.lib.table import STATUS_DOWNLOADING, STATUS_NOT_DOWNLOAD, Download, Session
from bgmi.plugin.download import BaseDownloadService
from bgmi.utils import bangumi_save_path, print_error, print_info
from bgmi.website.base import Episode


def get_download_driver(delegate: str) -> BaseDownloadService:
    try:
        return cast(
            BaseDownloadService,
            stevedore.DriverManager(namespace.DOWNLOAD_DELEGATE, name=delegate, invoke_on_load=True).driver,
        )
    except NoMatches:
        print_error(f"can't load download delegate {delegate}")
        raise


def download_episodes(data: List[Episode]) -> None:
    driver = get_download_driver(cfg.download_delegate)

    for episode in data:
        save_path = bangumi_save_path(episode.name).joinpath(str(episode.episode))
        if not save_path.exists():
            os.makedirs(save_path)

        download = Download(
            title=episode.title,
            download=episode.download,
            bangumi_name=episode.name,
            episode=episode.episode,
            status=STATUS_DOWNLOADING,
        )

        with Session.begin() as session:
            session.add(download)

        try:
            driver.add_download(url=download.download, save_path=str(save_path))
            print_info("Add torrent into the download queue, " f"the file will be saved at {save_path}")
        except Exception as e:
            if os.getenv("DEBUG"):  # pragma: no cover
                traceback.print_exc()
                raise e

            print_error(f"Error when downloading {download.title}: {e}", stop=False)
            download.status = STATUS_NOT_DOWNLOAD
            download.save()
